==> copilot-24.5.7-sources.jar: com/vaadin/copilot/ComponentSourceFinder.java <==
package com.vaadin.copilot;

import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.atomic.AtomicReference;

import com.vaadin.copilot.javarewriter.ComponentTypeAndSourceLocation;
import com.vaadin.flow.component.Component;
import com.vaadin.flow.component.internal.ComponentTracker;
import com.vaadin.flow.dom.Element;
import com.vaadin.flow.server.VaadinSession;

import elemental.json.JsonObject;

/**
 * Finds the source location of a component in the project.
 */
public class ComponentSourceFinder {

    private final ProjectManager projectManager;

    /**
     * Creates a new instance of the finder.
     *
     * @param projectManager
     *            the project manager to use
     */
    public ComponentSourceFinder(ProjectManager projectManager) {
        this.projectManager = projectManager;
    }

    /**
     * Finds the source location of a component.
     *
     * @param component
     *            the component to find, defined by uiId and nodeId in the given
     *            JSON object
     * @return the source location of the component
     */
    public ComponentTypeAndSourceLocation findTypeAndSourceLocation(JsonObject component) {
        return findTypeAndSourceLocation(component, false);
    }

    /**
     * Finds the source location of a component.
     *
     * @param component
     *            the component to find, defined by uiId and nodeId in the given
     *            JSON object
     * @return the source location of the component
     */
    public ComponentTypeAndSourceLocation findTypeAndSourceLocation(JsonObject component, boolean includeChildren) {
        return findTypeAndSourceLocation((int) component.getNumber("uiId"), (int) component.getNumber("nodeId"),
                includeChildren);
    }

    /**
     * Finds the source location of a component.
     *
     * @param uiId
     *            the uiId of the component
     * @param nodeId
     *            the nodeId of the component
     * @return the source location of the component
     */
    public ComponentTypeAndSourceLocation findTypeAndSourceLocation(int uiId, int nodeId) {
        return findTypeAndSourceLocation(uiId, nodeId, false);
    }

    /**
     * Finds the source location of a component.
     *
     * @param uiId
     *            the uiId of the component
     * @param nodeId
     *            the nodeId of the component
     * @param includeChildren
     *            whether to include children in the search
     * @return the source location of the component
     */
    public ComponentTypeAndSourceLocation findTypeAndSourceLocation(int uiId, int nodeId, boolean includeChildren) {
        AtomicReference<ComponentTypeAndSourceLocation> info = new AtomicReference<>(null);

        VaadinSession vaadinSession = projectManager.getVaadinSession();
        vaadinSession.accessSynchronously(() -> {
            Element element = vaadinSession.findElement(uiId, nodeId);
            Optional<Component> c = element.getComponent();
            if (c.isPresent()) {
                ComponentTypeAndSourceLocation componentInfo = findTypeAndSourceLocation(c.get(), includeChildren);
                info.set(componentInfo);
            }
        });
        ComponentTypeAndSourceLocation res = info.get();
        if (res == null) {
            throw new IllegalArgumentException("Unable to find component in source");
        }
        return res;
    }

    /**
     * Finds the source location of a component.
     *
     * @param component
     *            the component to find
     * @param includeChildren
     *            whether to include children in the search
     * @return the source location of the component
     */
    public ComponentTypeAndSourceLocation findTypeAndSourceLocation(Component component, boolean includeChildren) {
        Optional<Component> parent = component.getParent();

        ComponentTypeAndSourceLocation parentInfo = parent.map(this::_getSourceLocation).orElse(null);

        List<ComponentTypeAndSourceLocation> children = new ArrayList<>();
        if (includeChildren) {
            component.getChildren().forEach(child -> {
                ComponentTypeAndSourceLocation childInfo = findTypeAndSourceLocation(child, true);
                children.add(childInfo);
            });
        }
        return getSourceLocation(component, parentInfo, children);
    }

    private record LocationAndFile(ComponentTracker.Location location, File file) {
    }

    /**
     * Gets the source location for the given component.
     * <p>
     * NOTE that the session must be locked when calling this method.
     *
     * @return the source location of the component
     */
    public ComponentTypeAndSourceLocation _getSourceLocation(Component component) {
        return getSourceLocation(component, null, null);
    }

    private ComponentTypeAndSourceLocation getSourceLocation(Component component,
            ComponentTypeAndSourceLocation parentInfo, List<ComponentTypeAndSourceLocation> children) {
        ComponentTracker.Location[] createLocations = ComponentTracker.findCreateLocations(component);
        ComponentTracker.Location[] attachLocations = ComponentTracker.findAttachLocations(component);
        Optional<LocationAndFile> create = findProjectLocation(createLocations);
        Optional<LocationAndFile> attach = findProjectLocation(attachLocations);
        Optional<ComponentTracker.Location> createLocation = create.map(LocationAndFile::location);
        Optional<ComponentTracker.Location> attachLocation = attach.map(LocationAndFile::location);
        Optional<File> javaFile = createLocation.map(projectManager::getSourceFile);

        return new ComponentTypeAndSourceLocation(component.getClass(), findInheritanceChain(component), component,
                javaFile.orElse(null), createLocation, attachLocation, parentInfo, children);
    }

    /**
     * Finds inheritance chain of a component by going up for each superclass
     *
     * @param component
     *            Flow component
     * @return list of classes
     */
    public static List<Class<?>> findInheritanceChain(Component component) {
        List<Class<?>> inheritanceChain = new ArrayList<>();
        Class<?> superClass = component.getClass().getSuperclass();
        inheritanceChain.add(component.getClass());
        while (superClass != null) {
            inheritanceChain.add(superClass);
            superClass = superClass.getSuperclass();
        }

        return inheritanceChain;
    }

    private Optional<LocationAndFile> findProjectLocation(ComponentTracker.Location[] locations) {
        if (locations == null || locations.length == 0) {
            return Optional.empty();
        }
        for (ComponentTracker.Location location : locations) {
            File file = ProjectFileManager.getInstance().getSourceFile(location);
            if (file.exists()) {
                return Optional.of(new LocationAndFile(location, file));
            }
        }

        return Optional.empty();
    }
}

==> copilot-24.5.10-sources.jar: com/vaadin/copilot/ComponentSourceFinder.java <==
package com.vaadin.copilot;

import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.atomic.AtomicReference;
import java.util.stream.Stream;

import com.vaadin.copilot.javarewriter.ComponentTypeAndSourceLocation;
import com.vaadin.flow.component.Component;
import com.vaadin.flow.component.internal.ComponentTracker;
import com.vaadin.flow.dom.Element;
import com.vaadin.flow.server.VaadinSession;

import elemental.json.JsonObject;

/**
 * Finds the source location of a component in the project.
 */
public class ComponentSourceFinder {

    private static final String[] prefixesToSkip = new String[] { "com.vaadin.flow.component.internal.",
            "com.vaadin.flow.di.", "com.vaadin.flow.dom.", "com.vaadin.flow.internal.", "com.vaadin.flow.spring.",
            "com.vaadin.cdi.", "java.", "jdk.", "org.springframework.beans.", "org.jboss.weld.", };

    private final ProjectManager projectManager;

    private enum Type {
        CREATE, ATTACH
    }

    /**
     * Creates a new instance of the finder.
     *
     * @param projectManager
     *            the project manager to use
     */
    public ComponentSourceFinder(ProjectManager projectManager) {
        this.projectManager = projectManager;
    }

    /**
     * Finds the source location of a component.
     *
     * @param component
     *            the component to find, defined by uiId and nodeId in the given
     *            JSON object
     * @return the source location of the component
     */
    public ComponentTypeAndSourceLocation findTypeAndSourceLocation(JsonObject component) {
        return findTypeAndSourceLocation(component, false);
    }

    /**
     * Finds the source location of a component.
     *
     * @param component
     *            the component to find, defined by uiId and nodeId in the given
     *            JSON object
     * @return the source location of the component
     */
    public ComponentTypeAndSourceLocation findTypeAndSourceLocation(JsonObject component, boolean includeChildren) {
        return findTypeAndSourceLocation((int) component.getNumber("uiId"), (int) component.getNumber("nodeId"),
                includeChildren);
    }

    /**
     * Finds the source location of a component.
     *
     * @param uiId
     *            the uiId of the component
     * @param nodeId
     *            the nodeId of the component
     * @return the source location of the component
     */
    public ComponentTypeAndSourceLocation findTypeAndSourceLocation(int uiId, int nodeId) {
        return findTypeAndSourceLocation(uiId, nodeId, false);
    }

    /**
     * Finds the source location of a component.
     *
     * @param uiId
     *            the uiId of the component
     * @param nodeId
     *            the nodeId of the component
     * @param includeChildren
     *            whether to include children in the search
     * @return the source location of the component
     */
    public ComponentTypeAndSourceLocation findTypeAndSourceLocation(int uiId, int nodeId, boolean includeChildren) {
        AtomicReference<ComponentTypeAndSourceLocation> info = new AtomicReference<>(null);

        VaadinSession vaadinSession = projectManager.getVaadinSession();
        vaadinSession.accessSynchronously(() -> {
            Element element = vaadinSession.findElement(uiId, nodeId);
            Optional<Component> c = element.getComponent();
            if (c.isPresent()) {
                ComponentTypeAndSourceLocation componentInfo = findTypeAndSourceLocation(c.get(), includeChildren);
                info.set(componentInfo);
            }
        });
        ComponentTypeAndSourceLocation res = info.get();
        if (res == null) {
            throw new IllegalArgumentException("Unable to find component in source");
        }
        return res;
    }

    /**
     * Finds the source location of a component.
     *
     * @param component
     *            the component to find
     * @param includeChildren
     *            whether to include children in the search
     * @return the source location of the component
     */
    public ComponentTypeAndSourceLocation findTypeAndSourceLocation(Component component, boolean includeChildren) {
        Optional<Component> parent = component.getParent();

        ComponentTypeAndSourceLocation parentInfo = parent.map(this::_getSourceLocation).orElse(null);

        List<ComponentTypeAndSourceLocation> children = new ArrayList<>();
        if (includeChildren) {
            component.getChildren().forEach(child -> {
                ComponentTypeAndSourceLocation childInfo = findTypeAndSourceLocation(child, true);
                children.add(childInfo);
            });
        }
        return getSourceLocation(component, parentInfo, children);
    }

    private record LocationAndFile(ComponentTracker.Location location, File file) {
    }

    /**
     * Gets the source location for the given component.
     * <p>
     * NOTE that the session must be locked when calling this method.
     *
     * @return the source location of the component
     */
    public ComponentTypeAndSourceLocation _getSourceLocation(Component component) {
        return getSourceLocation(component, null, null);
    }

    private ComponentTypeAndSourceLocation getSourceLocation(Component component,
            ComponentTypeAndSourceLocation parentInfo, List<ComponentTypeAndSourceLocation> children) {
        ComponentTracker.Location[] createLocations = ComponentTracker.findCreateLocations(component);
        ComponentTracker.Location[] attachLocations = ComponentTracker.findAttachLocations(component);
        Optional<LocationAndFile> create = findProjectLocation(component, createLocations, Type.CREATE);
        Optional<LocationAndFile> attach = findProjectLocation(component, attachLocations, Type.ATTACH);
        Optional<ComponentTracker.Location> createLocation = create.map(LocationAndFile::location);
        Optional<ComponentTracker.Location> attachLocation = attach.map(LocationAndFile::location);
        Optional<File> javaFile = createLocation.map(projectManager::getSourceFile);

        if (javaFile.isEmpty()) {
            // The component was not created in the project source
            createLocation = Optional.empty();
            attachLocation = Optional.empty();
        }
        return new ComponentTypeAndSourceLocation(component.getClass(), findInheritanceChain(component), component,
                javaFile, createLocation, attachLocation, parentInfo, children);
    }

    /**
     * Finds inheritance chain of a component by going up for each superclass
     *
     * @param component
     *            Flow component
     * @return list of classes
     */
    public static List<Class<?>> findInheritanceChain(Component component) {
        List<Class<?>> inheritanceChain = new ArrayList<>();
        Class<?> superClass = component.getClass().getSuperclass();
        inheritanceChain.add(component.getClass());
        while (superClass != null) {
            inheritanceChain.add(superClass);
            superClass = superClass.getSuperclass();
        }

        return inheritanceChain;
    }

    private Optional<LocationAndFile> findProjectLocation(Component component, ComponentTracker.Location[] allLocations,
            Type type) {
        if (allLocations == null) {
            return Optional.empty();
        }

        List<ComponentTracker.Location> locations = Stream.of(allLocations).filter(location -> {
            for (String prefixToSkip : prefixesToSkip) {
                if (location.className().startsWith(prefixToSkip)) {
                    return false;
                }
            }
            return true;
        }).toList();

        // First find the <init> for the given component
        for (int i = 0; i < locations.size(); i++) {
            ComponentTracker.Location location = locations.get(i);
            if (type == Type.CREATE && isConstructor(location, component)) {

                int createIndex = i;
                // Internal constructor calls are not interesting
                while (isConstructor(locations.get(createIndex + 1), component)) {
                    createIndex++;
                }

                // This is the constructor call for the component with are looking for
                // The next stack trace is the place where the component is created (unless
                // there are some intermediate proxy layers like with Spring)
                // However, if the "intermediate" layers contain a component creation, we cannot
                // look at those as we would pick up the location where the parent component is
                // created instead
                Optional<LocationAndFile> componentCreate = findInProject(locations.get(createIndex + 1));

                if (componentCreate.isPresent()) {
                    return componentCreate;
                }
                // This could be a route class, then we return the init of the class itself
                return findInProject(locations.get(i));
            } else if (type == Type.ATTACH) {
                Optional<LocationAndFile> locationAndFileInProject = findInProject(location);
                if (locationAndFileInProject.isPresent()) {
                    return locationAndFileInProject;
                }
            }
        }

        return Optional.empty();
    }

    private static boolean isConstructor(ComponentTracker.Location location, Component component) {
        return location.className().equals(component.getClass().getName()) && "<init>".equals(location.methodName());
    }

    private Optional<LocationAndFile> findInProject(ComponentTracker.Location location) {
        File file = ProjectFileManager.getInstance().getSourceFile(location);
        if (file.exists()) {
            return Optional.of(new LocationAndFile(location, file));
        }
        return Optional.empty();
    }
}
